package eu.pozimski.alarmclockconfigurator.http

import io.ktor.client.*
import io.ktor.client.call.*
import io.ktor.client.request.*
import io.ktor.client.statement.*
import io.ktor.http.*
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import kotlinx.datetime.DayOfWeek
import kotlinx.datetime.LocalDate
import kotlinx.datetime.toJavaLocalDate
import java.time.format.DateTimeFormatter

class WakeupClient(private val client: HttpClient, private val apiUrl: String) {
    companion object {
        const val endpoint = "/v1/wakeup/"
        val formatter: DateTimeFormatter = DateTimeFormatter.ofPattern("yyyyMMdd")
        val weekDays = listOf("", "MON", "TUE", "WED", "THU", "FRI", "SAT", "SUN")
    }

    suspend fun fetchWakeup(date: LocalDate): WakeupTime? {
        val formattedDate = formatter.format(date.toJavaLocalDate())
        return extractWakeupTime(request(formattedDate, HttpMethod.Get))
    }

    suspend fun fetchWakeup(dayOfWeek: DayOfWeek): WakeupTime? {
        val weekday = weekDays[dayOfWeek.value]
        return extractWakeupTime(request(weekday, HttpMethod.Get))
    }

    private suspend fun extractWakeupTime(httpResponse: HttpResponse): WakeupTime? {
        if (httpResponse.status == HttpStatusCode.OK) {
            return httpResponse.body()
        } else if (httpResponse.status == HttpStatusCode.NotFound) {
            return null
        }
        throw Exception("Unexpected http status code: " + httpResponse.status.value)
    }

    private suspend fun request(
        identifier: String,
        httpMethod: HttpMethod,
        wakeupTime: WakeupTime?
    ): HttpResponse {
        val url = apiUrl + endpoint + identifier
        return client.request(url) {
            method = httpMethod
            contentType(ContentType.Application.Json)
            if (httpMethod != HttpMethod.Get && wakeupTime != null) {
                setBody(wakeupTime)
            }
        }
    }

    private suspend fun request(
        identifier: String,
        httpMethod: HttpMethod
    ): HttpResponse {
        return request(identifier, httpMethod, null)
    }

    private suspend fun evaluateResponse(httpResponse: HttpResponse) {
        if (httpResponse.status != HttpStatusCode.OK) {
            throw Exception("Error while performing request " + httpResponse.bodyAsText())
        }
    }

    suspend fun writeWakeup(date: LocalDate, wakeupTime: WakeupTime) {
        val formattedDate = formatter.format(date.toJavaLocalDate())
        evaluateResponse(request(formattedDate, HttpMethod.Put, wakeupTime))
    }

    suspend fun writeWakeup(dayOfWeek: DayOfWeek, wakeupTime: WakeupTime) {
        val weekday = weekDays[dayOfWeek.value]
        return evaluateResponse(request(weekday, HttpMethod.Put, wakeupTime))
    }

    suspend fun deactivateWakeup(date: LocalDate) {
        val formattedDate = formatter.format(date.toJavaLocalDate())
        return evaluateResponse(request(formattedDate, HttpMethod.Put, WakeupTime(null, null)))
    }

    suspend fun deactivateWakeup(dayOfWeek: DayOfWeek) {
        val weekday = weekDays[dayOfWeek.value]
        return evaluateResponse(request(weekday, HttpMethod.Put, WakeupTime(null, null)))
    }

    suspend fun deleteWakeup(date: LocalDate) {
        val formattedDate = formatter.format(date.toJavaLocalDate())
        evaluateResponse(request(formattedDate, HttpMethod.Delete ))
    }

    suspend fun deleteWakeup(dayOfWeek: DayOfWeek) {
        val weekday = weekDays[dayOfWeek.value]
        evaluateResponse(request(weekday, HttpMethod.Delete ))
    }

}