package eu.pozimski.alarmclockconfigurator.ui.elements

import android.os.Build
import androidx.annotation.RequiresApi
import androidx.compose.foundation.*
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.aspectRatio
import androidx.compose.foundation.layout.padding
import androidx.compose.material.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.unit.dp
import com.google.accompanist.swiperefresh.SwipeRefresh
import com.google.accompanist.swiperefresh.rememberSwipeRefreshState
import eu.pozimski.alarmclockconfigurator.http.WakeupTime
import eu.pozimski.alarmclockconfigurator.http.WakeupTimeCache
import eu.pozimski.alarmclockconfigurator.viewmodels.WakeupTimeViewModel
import io.github.boguszpawlowski.composecalendar.SelectableCalendar
import io.github.boguszpawlowski.composecalendar.day.DayState
import io.github.boguszpawlowski.composecalendar.selection.DynamicSelectionState
import io.ktor.util.collections.*
import kotlinx.datetime.LocalDate
import java.time.DayOfWeek
import java.time.format.DateTimeFormatter

@RequiresApi(Build.VERSION_CODES.TIRAMISU)
@Composable
fun CalendarView(localDateCache: WakeupTimeCache<LocalDate>) {
    val calendarViewModels: MutableMap<java.time.LocalDate, WakeupTimeViewModel<LocalDate>> =
        remember {
            ConcurrentMap()
        }
    var refreshing by remember {
        mutableStateOf(false)
    }

    SwipeRefresh(state = rememberSwipeRefreshState(refreshing), onRefresh = {
        refreshing = true
        calendarViewModels.values.forEach { viewModel -> viewModel.refresh() }
        refreshing = false
    }) {
        SelectableCalendar(
            firstDayOfWeek = DayOfWeek.MONDAY,
            showAdjacentMonths = false,
            dayContent = {
                WakeupDay(
                    it,
                    calendarViewModels.computeIfAbsent(
                        it.date
                    ) { date: java.time.LocalDate ->
                        WakeupTimeViewModel(
                            localDateCache,
                            LocalDate.fromEpochDays(
                                date.toEpochDay().toInt()
                            )
                        )
                    })
            },
            horizontalSwipeEnabled = false,
            modifier = Modifier.verticalScroll(rememberScrollState())
        )
    }
}

@RequiresApi(Build.VERSION_CODES.TIRAMISU)
@OptIn(ExperimentalFoundationApi::class)
@Composable
fun WakeupDay(
    state: DayState<DynamicSelectionState>,
    viewModel: WakeupTimeViewModel<LocalDate>
) {

    val date = state.date
    val selectionState = state.selectionState

    val isSelected = selectionState.isDateSelected(date)

    var dropDownExpanded by remember { mutableStateOf(false) }
    var showDeleteAlertDialog by remember { mutableStateOf(false) }
    var showDisableAlertDialog by remember { mutableStateOf(false) }
    var showTimePickerDialog by remember {
        mutableStateOf(false)
    }
    var currentError by remember {
        mutableStateOf<String?>(null)
    }
    val backGroundColor by viewModel.backGroundColor.collectAsState()

    Card(
        modifier = Modifier
            .aspectRatio(1f)
            .padding(2.dp),
        elevation = if (state.isFromCurrentMonth) 4.dp else 0.dp,
        border = if (state.isCurrentDay) BorderStroke(1.dp, MaterialTheme.colors.primary) else null,
        contentColor = if (isSelected) MaterialTheme.colors.secondary else contentColorFor(
            backgroundColor = MaterialTheme.colors.surface
        ),
        backgroundColor = backGroundColor
    ) {
        Box(
            modifier = Modifier.combinedClickable(onClick = {
                selectionState.onDateSelected(date)
                showTimePickerDialog = true
            },
                onLongClick = { dropDownExpanded = true }),
            contentAlignment = Alignment.Center,
        ) {
            Text(text = date.dayOfMonth.toString())
        }
    }

    if (showTimePickerDialog) {
        TimePicker(
            hour = viewModel.hour,
            minute = viewModel.minute,
            onTimeSetListener = { _, newHour: Int, newMinute: Int ->
                showTimePickerDialog = false
                currentError = viewModel.setWakeupTime(WakeupTime(newHour, newMinute))
            },
            onDismissListener = { showTimePickerDialog = false })
    }

    DropdownMenu(expanded = dropDownExpanded, onDismissRequest = { dropDownExpanded = false }) {
        DropdownMenuItem(onClick = { showDisableAlertDialog = true }) {
            Text("Disable")
        }
        DropdownMenuItem(onClick = { showDeleteAlertDialog = true }) {
            Text("Delete")
        }
    }

    if (currentError != null) {
        ErrorDialog(onDismissRequest = { currentError = null}, currentError)
    }

    if (showDeleteAlertDialog) {
        DeleteAlertDialog(onDismissRequest = {
            showDeleteAlertDialog = false
            dropDownExpanded = false
        }, onConfirm = {
            showDeleteAlertDialog = false
            dropDownExpanded = false
            currentError = viewModel.deleteWakeupTime()
        }, deletionElementNameCreator = {
            date.format(
                DateTimeFormatter.ISO_DATE
            )
        })
    }
    if (showDisableAlertDialog) {
        DisableAlertDialog(onDismissRequest = {
            showDisableAlertDialog = false
            dropDownExpanded = false
        }, onConfirm = {
            showDisableAlertDialog = false
            dropDownExpanded = false
            currentError = viewModel.deactivateWakeupTime()
        },
            elementNameCreator = {
                date.format(
                    DateTimeFormatter.ISO_DATE
                )
            })
    }
}