package eu.pozimski.alarmclockconfigurator.http

import io.ktor.client.*
import io.ktor.client.engine.mock.*
import io.ktor.client.plugins.contentnegotiation.*
import io.ktor.http.*
import io.ktor.serialization.kotlinx.json.*
import kotlinx.coroutines.runBlocking
import kotlinx.datetime.LocalDate
import org.junit.Assert.assertEquals
import org.junit.Assert.assertNotNull
import org.junit.Test
import java.time.DayOfWeek

class WakeupClientTest {

    @Test
    fun fetchWakeup_givenDateAndServerRepliesWithWakeupTime_returnsWakeupObject() {
        given()
            .serverRespondsWith("{\"hour\":5, \"minute\":0}", HttpStatusCode.OK)
            .assertThatCallingFetchFunction { wakeupClient ->
                wakeupClient.fetchWakeup(
                    LocalDate(
                        2023,
                        1,
                        2
                    )
                )
            }
            .callsUrl("http://localhost/v1/wakeup/20230102", HttpMethod.Get)
            .returnsWakeupTime(WakeupTime(5, 0))
    }

    @Test
    fun fetchWakeup_givenDateAndServerRepliesWithEmptyJsonObject_returnsWakeupObjectWithNullValues() {
        given()
            .serverRespondsWith("{}", HttpStatusCode.OK)
            .assertThatCallingFetchFunction { wakeupClient ->
                wakeupClient.fetchWakeup(
                    LocalDate(
                        2023,
                        1,
                        2
                    )
                )
            }
            .callsUrl("http://localhost/v1/wakeup/20230102", HttpMethod.Get)
            .returnsWakeupTime(WakeupTime(null, null))
    }

    @Test
    fun fetchWakeup_givenDateAndServerRepliesWithHttpStatusNotFound_returnsNull() {
        given()
            .serverRespondsWith("", HttpStatusCode.NotFound)
            .assertThatCallingFetchFunction { wakeupClient ->
                wakeupClient.fetchWakeup(
                    LocalDate(
                        2023,
                        1,
                        2
                    )
                )
            }
            .callsUrl("http://localhost/v1/wakeup/20230102", HttpMethod.Get)
            .returnsWakeupTime(null)
    }

    @Test
    fun fetchWakeup_givenDayOfWeekAndServerRepliesWithWakeupTime_returnsWakeupObject() {
        given()
            .serverRespondsWith("{\"hour\":5, \"minute\":0}", HttpStatusCode.OK)
            .assertThatCallingFetchFunction { wakeupClient -> wakeupClient.fetchWakeup(DayOfWeek.FRIDAY) }
            .callsUrl("http://localhost/v1/wakeup/FRI", HttpMethod.Get)
            .returnsWakeupTime(WakeupTime(5, 0))
    }

    @Test
    fun fetchWakeup_givenDayOfWeekAndServerRepliesWithEmptyJsonObject_returnsWakeupObjectWithNullValues() {
        given()
            .serverRespondsWith("{}", HttpStatusCode.OK)
            .assertThatCallingFetchFunction { wakeupClient -> wakeupClient.fetchWakeup(DayOfWeek.FRIDAY) }
            .callsUrl("http://localhost/v1/wakeup/FRI", HttpMethod.Get)
            .returnsWakeupTime(WakeupTime(null, null))
    }

    @Test
    fun fetchWakeup_givenDayOfWeekAndServerRepliesWithHttpStatusNotFound_returnsNull() {
        given()
            .serverRespondsWith("", HttpStatusCode.NotFound)
            .assertThatCallingFetchFunction { wakeupClient -> wakeupClient.fetchWakeup(DayOfWeek.FRIDAY) }
            .callsUrl("http://localhost/v1/wakeup/FRI", HttpMethod.Get)
            .returnsWakeupTime(null)
    }

    @Test
    fun fetchWakeup_givenDayOfWeekAndServerRepliesWithHttpStatusInternalServerError_throwsException() {
        given()
            .serverRespondsWith("", HttpStatusCode.InternalServerError)
            .assertThatCallingFetchFunction { wakeupClient -> wakeupClient.fetchWakeup(DayOfWeek.FRIDAY) }
            .callsUrl("http://localhost/v1/wakeup/FRI", HttpMethod.Get)
            .throwsException()
    }

    @Test
    fun writeWakeup_givenDateAndServerRepliesWithHttpStatusOK_performsPutRequestToCorrectUrlWithJsonObjectContainingWakeupTime() {
        given()
            .serverRespondsWith("", HttpStatusCode.OK)
            .assertThatCallingWriteFunction { wakeupClient ->
                wakeupClient.writeWakeup(
                    LocalDate(
                        2023,
                        1,
                        2
                    ), WakeupTime(5, 0)
                )
            }
            .callsUrl("http://localhost/v1/wakeup/20230102", HttpMethod.Put)
            .withJsonBody("{\"hour\":5,\"minute\":0}")
    }

    @Test
    fun writeWakeup_givenWeekDayAndServerRepliesWithHttpStatusOK_performsPutRequestToCorrectUrlWithJsonObjectContainingWakeupTime() {
        given()
            .serverRespondsWith("", HttpStatusCode.OK)
            .assertThatCallingWriteFunction { wakeupClient ->
                wakeupClient.writeWakeup(
                    DayOfWeek.FRIDAY,
                    WakeupTime(5, 0)
                )
            }
            .callsUrl("http://localhost/v1/wakeup/FRI", HttpMethod.Put)
            .withJsonBody("{\"hour\":5,\"minute\":0}")
    }

    @Test
    fun deactivateWakeup_givenDateAndServerRepliesWithHttpStatusOK_performsPutRequestToCorrectUrlWithEmptyJsonObject() {
        given()
            .serverRespondsWith("", HttpStatusCode.OK)
            .assertThatCallingWriteFunction { wakeupClient ->
                wakeupClient.deactivateWakeup(
                    LocalDate(2023, 1, 2)
                )
            }
            .callsUrl("http://localhost/v1/wakeup/20230102", HttpMethod.Put)
            .withJsonBody("{}")
    }

    @Test
    fun deactivateWakeup_givenWeekDayAndServerRepliesWithHttpStatusOK_performsPutRequestToCorrectUrlWithEmptyJsonObject() {
        given()
            .serverRespondsWith("", HttpStatusCode.OK)
            .assertThatCallingWriteFunction { wakeupClient ->
                wakeupClient.deactivateWakeup(
                    DayOfWeek.FRIDAY
                )
            }
            .callsUrl("http://localhost/v1/wakeup/FRI", HttpMethod.Put)
            .withJsonBody("{}")
    }

    @Test
    fun deleteWakeup_givenDateAndServerRepliesWithHttpStatusOK_performsDeleteRequestToCorrectUrl() {
        given()
            .serverRespondsWith("", HttpStatusCode.OK)
            .assertThatCallingWriteFunction { wakeupClient ->
                wakeupClient.deleteWakeup(
                    LocalDate(
                        2023,
                        1,
                        2
                    )
                )
            }
            .callsUrl("http://localhost/v1/wakeup/20230102", HttpMethod.Delete)
    }

    @Test
    fun deleteWakeup_givenDateAndServerRepliesWithHttpStatusInternalServerError_throwsException() {
        given()
            .serverRespondsWith("", HttpStatusCode.InternalServerError)
            .assertThatCallingWriteFunction { wakeupClient ->
                wakeupClient.deleteWakeup(
                    LocalDate(
                        2023,
                        1,
                        2
                    )
                )
            }
            .callsUrl("http://localhost/v1/wakeup/20230102", HttpMethod.Delete)
            .throwsException()
    }

    private fun given(): TestBuilder {
        return TestBuilder()
    }

    private class TestBuilder {
        private lateinit var mockEngine: MockEngine
        private var actual: WakeupTime? = null
        private lateinit var urlCalled: String
        private lateinit var httpMethod: HttpMethod
        private lateinit var thrownException: Exception
        private lateinit var requestBody: String

        fun serverRespondsWith(content: String, statusCode: HttpStatusCode): TestBuilder {
            mockEngine = MockEngine { request ->
                urlCalled = request.url.toString()
                httpMethod = request.method
                requestBody = request.body.toByteArray().decodeToString()
                respond(
                    content = content,
                    status = statusCode,
                    headers = headersOf(HttpHeaders.ContentType, "application/json")
                )
            }
            return this
        }

        fun assertThatCallingFetchFunction(functionToTest: suspend (WakeupClient) -> WakeupTime?): TestBuilder {
            val wakeupClient = WakeupClient(HttpClient(mockEngine) {
                install(
                    ContentNegotiation
                ) {
                    json()
                }
            }, "")
            try {
                actual = runBlocking { functionToTest.invoke(wakeupClient) }
            } catch (e: Exception) {
                thrownException = e
            }
            return this
        }

        fun assertThatCallingWriteFunction(functionToTest: suspend (WakeupClient) -> Unit): TestBuilder {
            val wakeupClient = WakeupClient(HttpClient(mockEngine) {
                install(
                    ContentNegotiation
                ) {
                    json()
                }
            }, "")
            try {
                runBlocking { functionToTest.invoke(wakeupClient) }
            } catch (e: Exception) {
                thrownException = e
            }
            return this
        }

        fun returnsWakeupTime(expected: WakeupTime?) {
            assertEquals(expected, actual)
        }

        fun callsUrl(url: String, httpMethod: HttpMethod): TestBuilder {
            assertEquals(url, urlCalled)
            assertEquals(this.httpMethod, httpMethod)
            return this
        }

        fun withJsonBody(expectedBody: String) {
            assertEquals(expectedBody, requestBody)
        }

        fun throwsException() {
            assertNotNull(thrownException)
        }
    }
}